# 'Fake' lightflowPM module for use with Python version other than 1.5
# Does however work with python1.5 too, nothing version specific is used(?).

import sys

# Derive a (hopefully correct) path to the python version used to run this script
# without using the os module since it needs the correct path.
# This is done so (almost) any python version can be used including 1.5.
# Works sofar with 1.5.2, 2.0.1, 2.1, 2.2.1
# Of course this will only work with the actual python executable,
# NOT for embedded python like with Blender itself.
# NOT USED IN LINUX, WHICH DEPENDS ON HAVING PYTHONPATH CORRECTLY SET (binary not located in main python dir.)
def constructFullPyPath():
	st = sys.executable
	i = len(st)-1
	pdelim = '\\'
	while (i!=-1) and st[i]!=pdelim:
		i = i - 1
	if i<=0:
		raise "Cannot construct PYTHONPATH"
	mainpath = st[:i]
	return [mainpath, mainpath+'\\DLLS', mainpath+'\\LIB', mainpath+'\\LIB\\LIB-TK', mainpath+'\\LIB\\plat-win']

# insert the constructed paths inbetween current directory and any other
# path directories so these are the first directories searched before anything else
if sys.platform=='win32':
	newpath = constructFullPyPath()
	sys.path = [sys.path[0]] + newpath + sys.path[1:]

# with the path now hopefully correct, import the other modules needed
import os, signal
from math import sin, cos
from struct import unpack

# matrix functions

# returns matrix identity
def MtxIdentity():
	# return identity
	return [	[1.0, 0.0, 0.0, 0.0],
			[0.0, 1.0, 0.0, 0.0],
			[0.0, 0.0, 1.0, 0.0],
			[0.0, 0.0, 0.0, 1.0]	]

# creates full 3D rotation matrix 
# rx, ry, rz angles in radians 
def euler2mat(rx, ry, rz): 
	A, B = sin(rx), cos(rx)
	C, D = sin(ry), cos(ry)
	E, F = sin(rz), cos(rz)
	AC, BC = A*C, B*C
	return	[	[D*F,       D*E,     -C,	0.0],
			[AC*F-B*E, AC*E+B*F, A*D,	0.0],
			[BC*F+A*E, BC*E-A*F, B*D,	0.0],
			[0.0,		0.0,	0.0,	1.0]	]

# returns transpose of 4x4 matrix
def transp4x4(m):
	r = [[0,0,0,0], [0,0,0,0], [0,0,0,0], [0,0,0,0]]
	for i in range(4):
		for j in range(4):
			r[i][j] = m[j][i]
	return r


# dummy Lightflow classes & functions that will make the parsing easy
# and completely automatic, works for almost ANY Lightflow python file
# Most classes contain the absolute minimum just to parse the code, nothing more

# minimal matrix3 class, different from LF
class matrix3:
	def __init__(self, mtx):
		# expects mtx to be a list of lists
		self.mtx = mtx
	def __repr__(self):
		# self.mtx is list of lists (matrix)
		# Lightflow matrix is the transposed matrix calculated here
		# Blender row-major -> Lightflow column-major = tranpose(blender_mtx)
		st = 'matrix3 '
		for rw in self.mtx:
			for cm in rw:
				st = st + ('%s ' % cm)
		return st

# minimal matrix4 class
class matrix4:
	def __init__(self, mtx=MtxIdentity()):
		# expects mtx to be rm, like blender matrices
		# which is of course not true for the real LF class, which also supports overloaded versions(float list, matrix, vector list)
		self.mtx = mtx
	def __repr__(self):
		# self.mtx is list of lists (matrix)
		# Lightflow matrix is the transposed matrix calculated here
		# Blender row-major -> Lightflow column-major = tranpose(blender_mtx)
		st = 'matrix4 '
		for rw in self.mtx:
			for cm in rw:
				st = st + ('%s ' % cm)
		return st
	

# BSpline basis matrices
def BSplineBasis():
	oos = 1.0/6.0
	tot = 2.0/3.0
	return matrix4(	[[-oos,	0.5,	-0.5,	oos],
			 [0.5,	-1.0,	0.5,	0.0],
			 [-0.5,	0.0,	0.5,	0.0],
			 [oos,	tot,	oos,	0.0]] )

def BezierBasis():
	return matrix4(	[[-1.0,	3.0,	-3.0,	1.0],
			 [3.0,	-6.0,	3.0,	0.0],
			 [-3.0,	3.0,	0.0,	0.0],
			 [1.0,	0.0,	0.0,	0.0]] )

def CatmullRomBasis():
	return matrix4(	[[-0.5,	1.5,	-1.5,	0.5],
			 [1.0,	-2.5,	2.0,	-0.5],
			 [-0.5,	0.0,	0.5,	0.0],
			 [0.0,	1.0,	0.0,	0.0]] )

def HermiteBasis():
	# all zero??????
	return matrix4(	[[0.0,	0.0,	0.0,	0.0],
			 [0.0,	0.0,	0.0,	0.0],
			 [0.0,	0.0,	0.0,	0.0],
			 [0.0,	0.0,	0.0,	0.0]] )


# BSpline step values
def BSplineStep():
	return 1

def BezierStep():
	return 3

def CatmullRomStep():
	return 1

def HermiteStep():
	return 2


# 2 element vector
class vector2:
	def __init__(self, a=0.0, b=0.0):
		self.xy = (a, b)
	def set(self, index, val):
		t = list(self.xy)
		t[index] = val
		self.xy = tuple(t)
	def get(self, idx):
		return self.xy[idx]
	def __repr__(self):
		return 'vector2 %s %s' % self.xy


# array of vector2
class vector2array:
	def __init__(self):
		self.v2a = []
	def setSize(self, size):
		for i in range(size):
			self.v2a.append(vector2())
	def getSize(self):
		return len(self.v2a)
	def set(self, index, v2):
		self.v2a[index] = v2
	def get(self, index):
		return self.v2a[index]
	def add(self, v2):
		self.v2a.append(v2)
	def __repr__(self):
		sz = len(self.v2a)
		st = 'vector2-array ' + str(sz) + ' '
		for i in range(sz):
			st = st + '%s %s ' % self.v2a[i].xy
		return st


# 3 element vector
class vector3:
	def __init__(self, a=0.0, b=0.0, c=0.0):
		self.xyz = (a, b, c)
		self.x = a
		self.y = b
		self.z = c
	def set(self, index, val):
		t = list(self.xyz)
		t[index] = val
		self.xyz = tuple(t)
	def get(self, idx):
		return self.xyz[idx]
	def __repr__(self):
		return 'vector3 %s %s %s' % self.xyz


# array of vector3
class vector3array:
	def __init__(self):
		self.v3a = []
	def setSize(self, size):
		for i in range(size):
			self.v3a.append(vector3())
	def getSize(self):
		return len(self.v3a)
	def set(self, index, v3):
		self.v3a[index] = v3
	def get(self, index):
		return self.v3a[index]
	def add(self, v3):
		self.v3a.append(v3)
	def __repr__(self):
		sz = len(self.v3a)
		st = 'vector3-array ' + str(sz) + ' '
		for i in range(sz):
			st = st + '%s %s %s ' % self.v3a[i].xyz
		return st

# 4 element vector
class vector4:
	def __init__(self, a=0.0, b=0.0, c=0.0, d=0.0):
		self.xyzw = (a, b, c, d)
		self.x = a
		self.y = b
		self.z = c
		self.w = d
	def set(self, index, val):
		t = list(self.xyzw)
		t[index] = val
		self.xyzw = tuple(t)
	def get(self, idx):
		return self.xyzw[idx]
	def __repr__(self):
		return 'vector4 %s %s %s %s' % self.xyzw


# array of vector4
class vector4array:
	def __init__(self):
		self.v4a = []
	def setSize(self, size):
		for i in range(size):
			self.v4a.append(vector4())
	def getSize(self):
		return len(self.v4a)
	def set(self, index, v4):
		self.v4a[index] = v4
	def get(self, index):
		return self.v4a[index]
	def add(self, v4):
		self.v4a.append(v4)
	def __repr__(self):
		sz = len(self.v4a)
		st = 'vector4-array ' + str(sz) + ' '
		for i in range(sz):
			st = st + '%s %s %s %s ' % self.v4a[i].xyzw
		return st


# array of integers
class intarray:
	def __init__(self):
		self.ia = []
	def setSize(self, size):
		for i in range(size):
			self.ia.append(0)
	def getSize(self):
		return len(self.ia)
	def set(self, index, i):
		self.ia[index] = i
	def get(self, index):
		return self.ia[index]
	def add(self, i):
		self.ia.append(i)
	def __repr__(self):
		sz = len(self.ia)
		st = 'int-array ' + str(sz) + ' '
		for i in range(sz):
			st = st + '%s ' % self.ia[i]
		return st


# array of floats
class floatarray:
	def __init__(self):
		self.fa = []
	def setSize(self, size):
		for i in range(size):
			self.fa.append(0.0)
	def getSize(self):
		return len(self.fa)
	def set(self, index, f):
		self.fa[index] = f
	def get(self, index):
		return self.fa[index]
	def add(self, f):
		self.fa.append(f)
	def __repr__(self):
		sz = len(self.fa)
		st = 'float-array ' + str(sz) + ' '
		for i in range(sz):
			st = st + '%s ' % self.fa[i]
		return st


# transform class, does not use matrix4
# matrices are in row-major format here
class transform:

	#def __init__(self):
	#	self.matrix = MtxIdentity()

	def translation(self, v3):
		# argument is vector3
		mtx = MtxIdentity()
		mtx[3][0] = v3.x
		mtx[3][1] = v3.y
		mtx[3][2] = v3.z
		return mtx
	
	def rotationAroundX(self, arg):
		return euler2mat(arg, 0.0, 0.0)

	def rotationAroundY(self, arg):
		return euler2mat(0.0, arg, 0.0)

	def rotationAroundZ(self, arg):
		return euler2mat(0.0, 0.0, arg)

	def scaling(self, v3):
		# argument is vector3
		mtx = MtxIdentity()
		mtx[0][0] = v3.x
		mtx[1][1] = v3.y
		mtx[2][2] = v3.z
		return mtx

	# not part of lightflow, set matrix directly
	# mtx is 4x4 matrix as a list of lists
	# row-major version, argument is Blender style matrix, translation in third row
	def setMatrix_rm(self, mtx):
		return mtx

	# column-major, Lightflow style, translation in third column
	def setMatrix_cm(self, mtx):
		return transp4x4(mtx)


class scene:

	def __init__(self, texmem=10240):
		# texmem where??
		print "Python file parse start..."
		self.asciistring	=	'Begin\n'
		self.curcamera		=	-1
		self.curimager		=	-1
		self.curinterface	=	-1
		self.curinterior	=	-1
		self.curlight		=	-1
		self.curmaterial	=	-1
		self.curobject		=	-1
		self.curpattern		=	-1
		self.curtexture		=	-1
		self.curtrimmer		=	-1
		# set signal handler, moved out of main loop, only needs init. here
		signal.signal(signal.SIGINT, handler)

	def _parselist(self, arglist):
		for arg in arglist:
			tp = type(arg).__name__
			# for python>=22 --->> type() changed
			if tp=='str': tp = 'string'
			if tp in ['string', 'float', 'int']:
				self.asciistring = self.asciistring + tp + ' ' + str(arg) + '\n'
			else:
				self.asciistring = self.asciistring + str(arg) + '\n'

	# newXXXXXX...
	def newCamera(self, typ, arglist):
		self.asciistring = self.asciistring + '\nNewCamera\n'+ typ + '\n' + str(len(arglist)) + '\n'
		self._parselist(arglist)
		self.curcamera = self.curcamera + 1
		return self.curcamera

	def newImager(self, typ, arglist):
		self.asciistring = self.asciistring + '\nNewImager\n'+ typ + '\n' + str(len(arglist)) + '\n'
		self._parselist(arglist)
		self.curimager = self.curimager + 1
		return self.curimager

	def newInterface(self, typ, arglist):
		self.asciistring = self.asciistring + '\nNewInterface\n' + typ + '\n' + str(len(arglist)) + '\n'
		self._parselist(arglist)
		self.curinterface = self.curinterface + 1
		return self.curinterface

	def newInterior(self, typ, arglist):
		self.asciistring = self.asciistring + '\nNewInterior\n' + typ + '\n' + str(len(arglist)) + '\n'
		self._parselist(arglist)
		self.curinterior = self.curinterior + 1
		return self.curinterior

	def newLight(self, typ, arglist):
		self.asciistring = self.asciistring + '\nNewLight\n' + typ + '\n' + str(len(arglist)) + '\n'
		self._parselist(arglist)
		self.curlight = self.curlight + 1
		return self.curlight

	def newMaterial(self, typ, arglist):
		self.asciistring = self.asciistring + '\nNewMaterial\n'+ typ + '\n' + str(len(arglist)) + '\n'
		self._parselist(arglist)
		self.curmaterial = self.curmaterial + 1
		return self.curmaterial

	def newObject(self, typ, arglist):
		self.asciistring = self.asciistring + '\nNewObject\n'+ typ + '\n' + str(len(arglist)) + '\n'
		self._parselist(arglist)
		self.curobject = self.curobject + 1
		return self.curobject

	def newPattern(self, typ, arglist):
		self.asciistring = self.asciistring + '\nNewPattern\n'+ typ + '\n' + str(len(arglist)) + '\n'
		self._parselist(arglist)
		self.curpattern = self.curpattern + 1
		return self.curpattern

	def newTexture(self, arg):
		self.asciistring = self.asciistring + '\nNewTexture\n'+ arg + '\n'
		self.curtexture = self.curtexture + 1
		return self.curtexture

	def newTrimmer(self, typ, arglist):
		self.asciistring = self.asciistring + '\nNewTrimmer\n'+ typ + '\n' + str(len(arglist)) + '\n'
		self._parselist(arglist)
		self.curtrimmer = self.curtrimmer + 1
		return self.curtrimmer

	# IMAGERS
	def imagerBegin(self, args):
		self.asciistring = self.asciistring + '\nImagerBegin ' + str(args) + '\n'

	def imagerEnd(self):
		self.asciistring = self.asciistring + '\nImagerEnd\n'

	# INTERIOR
	def interiorBegin(self, args):
		self.asciistring = self.asciistring + '\nInteriorBegin ' + str(args) + '\n'

	def interiorEnd(self):
		self.asciistring = self.asciistring + 'InteriorEnd\n'

	# LIGHT
	def lightOn(self, args):
		self.asciistring = self.asciistring + '\nLightOn ' + str(args) + '\n'

	def lightOff(self, args):
		self.asciistring = self.asciistring + '\nLightOff ' + str(args) + '\n'

	def lightBegin(self):
		self.asciistring = self.asciistring + '\nLightBegin\n'

	def lightEnd(self):
		self.asciistring = self.asciistring + '\nLightEnd\n'

	# MATERIAL
	def materialBegin(self, args):
		self.asciistring = self.asciistring + '\nMaterialBegin ' + str(args) + '\n'

	def materialEnd(self):
		self.asciistring = self.asciistring + '\nMaterialEnd\n'

	# ADDOBJECT
	def addObject(self, args):
		self.asciistring = self.asciistring + '\nAddObject ' + str(args) + '\n'

	# TRANSFORM
	def transformBegin(self, mtx):
		self.asciistring = self.asciistring + '\nTransformBegin '
		# matrix argument is expected to be list of lists
		# Lightflow matrix is the transposed matrix calculated here
		# Blender row-major -> Lightflow column-major = transpose(blender_mtx)
		mtx = transp4x4(mtx)
		st = ''
		for rw in mtx:
			for cm in rw:
				st = st + ('%s ' % cm)
		self.asciistring = self.asciistring + st

	def transformEnd(self):
		self.asciistring = self.asciistring + 'TransformEnd\n'

	# RENDER
	def radiosity(self):
		self.asciistring = self.asciistring + '\nRadiosity\n'

	def render(self, *args):
		self.asciistring = self.asciistring + '\nRender '
		for arg in args:
			self.asciistring = self.asciistring + str(arg)+' '
		# needs image frame params if not specified.
		# width & height can be default too, but unlikely,
		# just assume always specified
		if len(args)==3:
			self.asciistring = self.asciistring + '0 0 1 1'
		self.asciistring = self.asciistring + '\n\nEnd\n'
		# scene data collected, save to temporary file
		# Use LIGHTFLOWTEMP directory as tempdir instead of system tmp
		print "Parse done, writing temporary lfa file"
		temppath = os.path.join(os.environ['LIGHTFLOWTEMP'], 'templfa')
		fp = open(temppath, 'w')
		fp.write(self.asciistring)
		fp.close()
		# render & wait until done
		print "Starting Lightflow..."
		lfapath = os.path.join(os.environ['LIGHTFLOWPATH'], 'LFArender')
		os.spawnv(os.P_WAIT, lfapath, (lfapath, temppath))
		# delete temporary file
		print "Deleting temporary file\n\n"
		os.remove(temppath)
		# reset sighandler, one time use only
		signal.signal(signal.SIGINT, signal.default_int_handler)


# IMPORTANT !!! Have to use sighandler to be able
# to catch CTRL-C interrupts so the temporary file can be deleted

# ctrl-c signal handler
def handler(signum, frame):
	print "\nRender aborted! Deleting temporary file\n\n"
	temppath = os.path.join(os.environ['LIGHTFLOWTEMP'], 'templfa')
	os.remove(temppath)
	# reset sighandler, one time use only
	signal.signal(signal.SIGINT, signal.default_int_handler)
	raise KeyboardInterrupt, "Ctrl-C was pressed"

# CONVERTS MESHFILE TO LF PATCHES & TRIANGLES
# not really useful, intention was to be able to use real displacement mapping on meshes
# but will push polys apart with high depth settings.
# Can also take too long on large meshes, making it seem like it has crashed.
# Of course the lfa file will be quite large too....
class mesh_reconstruct:
	"""
		Mesh unpack/reconstruct class for Lightflow.

		This is used to reconstruct a polygon mesh using
		triangles and patches, this is done so the surface-engine
		can be used for real displacement mapping.
	"""
	

	def __init__(self, mesh_filename, debug=0):
		fp = open(mesh_filename, 'rb')
		self.data = fp.read()
		fp.close()
		self.debug = debug
		# unpack number of meshes here
		self.num_meshes = unpack('l', self.data[0:4])[0]
		self.offset = 4		# data read offset
		if self.debug:
			print "Data length: ", len(self.data)
			print "number of meshes: ", self.num_meshes


	def _get_header_data(self):
		"unpack per mesh header(s)"
		self.nverts = []
		self.nnorms = []
		self.nuv = []
		self.nfaces = []
		self.ntverts = []

		for i in range(self.num_meshes):
			self.nverts.append(unpack('l', self.data[self.offset:self.offset+4])[0])
			self.nnorms.append(unpack('l', self.data[self.offset+4:self.offset+8])[0])
			self.nuv.append(unpack('l', self.data[self.offset+8:self.offset+12])[0])
			self.nfaces.append(unpack('l', self.data[self.offset+12:self.offset+16])[0])
			self.ntverts.append(unpack('l', self.data[self.offset+16:self.offset+20])[0])

			if self.debug:
				print "\nMesh %d Header:" % (i+1)
				print "Vertices: ", self.nverts[i]
				print "Normals: ", self.nnorms[i]
				print "UV-Coords: ", self.nuv[i]
				print "Faces: ", self.nfaces[i]
				print "Total vertices: ", self.ntverts[i]

			self.offset = self.offset + 20		# += >py20 !!!


	def _unpack_data(self, meshnum):
		"unpack mesh data"

		self.vert_list = []
		self.norm_list = []
		self.texc_list = []
		self.vert_per_face = []

		#print "Vertices:"
		for i in range(self.nverts[meshnum]):
			x = unpack('f', self.data[self.offset:self.offset+4])[0]
			y = unpack('f', self.data[self.offset+4:self.offset+8])[0]
			z = unpack('f', self.data[self.offset+8:self.offset+12])[0]
			self.vert_list.append((x,y,z))
			self.offset = self.offset + 12

		#print "\nNormals:"
		for i in range(self.nnorms[meshnum]):
			x = unpack('f', self.data[self.offset:self.offset+4])[0]
			y = unpack('f', self.data[self.offset+4:self.offset+8])[0]
			z = unpack('f', self.data[self.offset+8:self.offset+12])[0]
			self.norm_list.append((x, y, z))
			self.offset = self.offset + 12

		#print "\nTexCoords:"
		for i in range(self.nuv[meshnum]):
			x = unpack('f', self.data[self.offset:self.offset+4])[0]
			y = unpack('f', self.data[self.offset+4:self.offset+8])[0]
			self.texc_list.append((x, y))
			self.offset = self.offset + 8

		#print "\nverts per face:"
		for i in range(self.nfaces[meshnum]):
			x = unpack('l', self.data[self.offset:self.offset+4])[0]
			self.vert_per_face.append(x)
			self.offset = self.offset + 4


	def reconstruct(self, material_names, scene):
		"reconstruct mesh from data, using Lightflow patches & triangles"

		# unpack header(s) first
		self._get_header_data()

		# now the mesh data itself
		for curmesh in range(self.num_meshes):

			# unpack data for this mesh
			self._unpack_data(curmesh)

			# unpack index lists
			cur_vert_idx = []
			cur_norm_idx = []
			cur_texc_idx = []
			#print "\nvertex indices per face:"
			for i in range(self.nfaces[curmesh]):
				vidx = []
				for vertcount in range(self.vert_per_face[i]):
					x = unpack('l', self.data[self.offset:self.offset+4])[0]
					vidx.append(x)
					self.offset = self.offset + 4
				cur_vert_idx.append(vidx)

			#print "\nnormal indices per face:"
			for i in range(self.nfaces[curmesh]):
				nidx = []
				for vertcount in range(self.vert_per_face[i]):
					x = unpack('l', self.data[self.offset:self.offset+4])[0]
					nidx.append(x)
					self.offset = self.offset + 4
				cur_norm_idx.append(nidx)

			#print "\ntexcoords indices per face:"
			for i in range(self.nfaces[curmesh]):
				tidx = []
				for vertcount in range(self.vert_per_face[i]):
					x = unpack('l', self.data[self.offset:self.offset+4])[0]
					tidx.append(x)
					self.offset = self.offset + 4
				cur_texc_idx.append(tidx)

			# All data for this mesh unpacked and ordered, now create the objects
			# all obects share material for this mesh
			scene.materialBegin(material_names[curmesh])
			surfarray = intarray()
			for i in range(self.nfaces[curmesh]):
				# vertex, normal, texcoord indices for this face
				vidx = cur_vert_idx[i]
				nidx = cur_norm_idx[i]
				tidx = cur_texc_idx[i]
				if self.vert_per_face[i]==3:
					# triangle
					# LFBUG!!!! Lightflow 'triangle' does NOT accept anything but 'points' parameter....
					surfarray.add( scene.newObject('triangle', [
						'points',	apply(vector3, self.vert_list[vidx[0]]),
									apply(vector3, self.vert_list[vidx[1]]),
									apply(vector3, self.vert_list[vidx[2]])
						#'normals',	apply(vector3, self.norm_list[nidx[0]]),
						#			apply(vector3, self.norm_list[nidx[1]]),
						#			apply(vector3, self.norm_list[nidx[2]]),
						#'texture',	apply(vector2, self.texc_list[tidx[0]]),
						#			apply(vector2, self.texc_list[tidx[1]]),
						#			apply(vector2, self.texc_list[tidx[2]])
						] ))
				else:
					# quads (patches). vertex order: p1-p2-p4-p3 --> indexlist: 0-1-3-2
					surfarray.add( scene.newObject('patch', [
						'points',	apply(vector3, self.vert_list[vidx[0]]),
									apply(vector3, self.vert_list[vidx[1]]),
									apply(vector3, self.vert_list[vidx[3]]),
									apply(vector3, self.vert_list[vidx[2]]),
						'normals',	apply(vector3, self.norm_list[nidx[0]]),
									apply(vector3, self.norm_list[nidx[1]]),
									apply(vector3, self.norm_list[nidx[3]]),
									apply(vector3, self.norm_list[nidx[2]]),
						'texture',	apply(vector2, self.texc_list[tidx[0]]),
									apply(vector2, self.texc_list[tidx[1]]),
									apply(vector2, self.texc_list[tidx[3]]),
									apply(vector2, self.texc_list[tidx[2]])
						] ))

			# add all surfaces to scene
			scene.addObject(scene.newObject('surface-engine', ['surfaces', surfarray, 'tolerance', 0.02, 0.1, 0.2, 'curvature', 0.02]))

			# end material specifier
			scene.materialEnd()

			if self.debug:
				print "offset, datalength: ", self.offset, len(self.data)